package dk.apaq.crud.jpa;

import dk.apaq.crud.core.BaseCompleteCrud;
import dk.apaq.filter.Filter;
import dk.apaq.filter.jpa.FilterTranslatorForJPA;
import dk.apaq.filter.limit.Limit;
import dk.apaq.filter.sort.Sorter;
import java.lang.reflect.Field;
import java.util.List;
import javax.persistence.EntityManager;
import javax.persistence.EntityTransaction;
import javax.persistence.Id;
import javax.persistence.Query;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 */
public class EntityManagerCrud<IDTYPE, BEANTYPE> extends BaseCompleteCrud<IDTYPE, BEANTYPE> {

    private static final Logger LOG = LoggerFactory.getLogger(EntityManagerCrud.class);
    private final EntityManager entityManager;
    private final Class<BEANTYPE> clazz;
    private boolean transactionsHandled = true;

    @Deprecated
    public interface EntityManagerCrudAssist<IDTYPE, BEANTYPE> {

        /**
         * Retrieves the entity class.
         */
        Class<BEANTYPE> getEntityClass();

        /**
         * Creates a new instance of the entity.
         * @return
         */
        BEANTYPE createInstance();

        IDTYPE getIdForEntity(BEANTYPE entity);
    }

    @Deprecated
    public EntityManagerCrud(EntityManager entityManager, EntityManagerCrudAssist<IDTYPE, BEANTYPE> assist) {
        this(entityManager, assist.getEntityClass());
    }
    
    public EntityManagerCrud(EntityManager entityManager, Class<BEANTYPE> clazz) {
        this.entityManager = entityManager;
        this.clazz = clazz;
        
        if(entityManager==null) { throw new NullPointerException("EntityManager was null"); }
        if(clazz==null) { throw new NullPointerException("Class was null"); }
        
    }

    public BEANTYPE read(IDTYPE id) {
        fireOnBeforeRead(id);
        BEANTYPE entity = doRead(id);
        fireOnRead(id, entity);
        return entity;
    }

    public List<IDTYPE> listIds(Filter filter, Sorter sorter, Limit limit) {
        Query q = FilterTranslatorForJPA.translate(entityManager, new String[]{getIdProperty()}, clazz, filter, sorter, limit);
        return q.getResultList();
    }
    
    public List<BEANTYPE> list(Filter filter, Sorter sorter, Limit limit) {
        Query q = FilterTranslatorForJPA.translate(entityManager, clazz, filter, sorter, limit);
        return q.getResultList();
    }

    public IDTYPE create() {
        BEANTYPE entity;
        try {
            entity = clazz.newInstance();
        } catch (Exception ex) {
            LOG.error("Unable to create instance of class " + clazz.getCanonicalName(), ex);
            return null;
        }
        return create(entity);
        
    }

    public <T extends BEANTYPE> IDTYPE create(final T entity) {
        return getIdForEntity(createAndRead(entity));
    }

    public <T extends BEANTYPE> T createAndRead(final T entity) {
        fireOnBeforeCreate(entity);
        final Object[] entityA = new Object[1];
        runInTransaction(new Runnable() {

            public void run() {
                entityA[0] = entityManager.merge(entity);
                entityManager.flush();
            }
        });

        IDTYPE id = getIdForEntity((BEANTYPE) entityA[0]);
        fireOnCreate(id, entity);
        return (T) entityA[0];
    }
    
    
    

    public BEANTYPE update(final BEANTYPE entity) {
        IDTYPE id = this.getIdForEntity(entity);
        fireOnBeforeUpdate(id, entity);
        runInTransaction(new Runnable() {

            public void run() {
                entityManager.merge(entity);
                entityManager.flush();
            }
        });
        entityManager.merge(entity);
        fireOnUpdate(id, entity);
        return entity;
    }

    public void delete(IDTYPE id) {
        fireOnBeforeDelete(id);
        final BEANTYPE entity = doRead(id);
        runInTransaction(new Runnable() {

            public void run() {
                entityManager.remove(entity);
                entityManager.flush();
            }
        });
        fireOnDelete(id);
    }

    private BEANTYPE doRead(IDTYPE id) {
        return entityManager.find(clazz, id);
    }

    private String getIdProperty() {
        Field field = getIdField();
        if (field != null) {
            return field.getName();
        }
        return null;
    }

    private IDTYPE getIdForEntity(final BEANTYPE entity) {
        Field field = getIdField();
        if(field==null) {
            LOG.warn("Unable to find Id annotated field in entity.");
            return null;
        }
        
        field.setAccessible(true);
        try {
            return (IDTYPE) field.get(entity);
        } catch (Exception ex) {
            LOG.error("Unable to retrieve value from Id field.", ex);
            return null;
        }
    }

    private Field getIdField() {
        Class currentClass = this.clazz;

        while (currentClass != null) {
            for (Field field : currentClass.getDeclaredFields()) {
                if (field.getAnnotation(Id.class) != null) {
                    return field;
                }
            }
            currentClass = currentClass.getSuperclass();
        }
        return null;
    }

    /**
     * Specifies whether the entity provider should handle transactions itself
     * or whether they should be handled outside (e.g. if declarative
     * transactions are used).
     *
     * @param transactionsHandled
     *            true to handle the transactions internally, false to rely on
     *            external transaction handling.
     */
    public void setTransactionsHandledByProvider(boolean transactionsHandled) {
        this.transactionsHandled = transactionsHandled;
    }

    /**
     * Returns whether the entity provider is handling transactions internally
     * (the default) or relies on external transaction handling.
     *
     * @return true if transactions are handled internally, false if not.
     */
    public boolean isTransactionsHandledByProvider() {
        return transactionsHandled;
    }

    /**
     * If {@link #isTransactionsHandledByProvider() } is true,
     * <code>operation</code> will be executed inside a transaction that is
     * commited after the operation is completed. Otherwise,
     * <code>operation</code> will just be executed.
     *
     * @param operation the operation to run (must not be null).
     */
    protected void runInTransaction(Runnable operation) {
        assert operation != null : "operation must not be null";
        if (isTransactionsHandledByProvider()) {
            EntityTransaction et = entityManager.getTransaction();
            if (et.isActive()) {
                // The transaction has been started outside of this method
                // and should also be committed/rolled back outside of
                // this method
                operation.run();
            } else {
                try {
                    et.begin();
                    operation.run();
                    et.commit();
                } finally {
                    if (et.isActive()) {
                        et.rollback();
                    }
                }
            }
        } else {
            operation.run();
        }
    }
}
